"""
Constants should be defined here.
"""

TOP_LEVEL_FOLDER_NAME = 'packages'
DRAFT_RELEASE_NAME = 'draft'
APPLICATION_PACKAGE_TEMPLATE_NAME = '{baseName}_{os}_{arch}_{revision}'
EXTENSION_PACKAGE_TEMPLATE_NAME = '{app_revision}_{baseName}_{os}_{arch}_{revision}'
EXTENSIONS_FOLDER_NAME = 'extensions'
